#!/bin/env perl
#
# Copyright (c) 2018 IBM Corporation.  All rights reserved.
#
use strict;
use Class::Struct;
use XML::Simple;
use Data::Dumper;
use Getopt::Long;

######################################################################

# Decoder ring for coll ID
my %decode_coll_id = (  0 => "Bcast",
                        1 => "Allreduce",
                        2 => "Reduce",
                        3 => "Allgather",
                        4 => "Allgatherv",
                        5 => "Allgatherv_int",
                        6 => "Scatter",
                        7 => "Scatterv",
                        8 => "Scatterv_int",
                        9 => "Gather",
                       10 => "Gatherv",
                       11 => "Gatherv_int",
                       12 => "Barrier",
                       13 => "Alltoall",
                       14 => "Alltoallv",
                       15 => "Alltoallv_int",
                       16 => "Scan",
                       17 => "Reduce_scatter"
    );

######################################################################
struct GeometrySelection => {
    size => '$', # Size of the geometry
    items => '@', # Array of Geometry Items
};

struct GeometryItem => {
    size_min => '$',
    size_max => '$',
    algorithms => '@',
};

struct CollectiveSet => {
    id => '$', # ID of the collective operation
    algs => '%', # Hash of all CollectiveAlg (key = 'id')
    geometry => '%', # Geometry selection data, (key = 'geometry_size')
};

struct CollectiveAlg => {
    id => '$', # ID of the collective algorithm
    name => '$', # String name of the collective algorithm
};

my %all_collectives;

######################################################################
my $help_arg;
my $show_coll;
my $list_algs;

&Getopt::Long::Configure("bundling");
my $ok = Getopt::Long::GetOptions("help|h" => \$help_arg,
                                  "collective|c=s" => \$show_coll,
                                  "list|l" => \$list_algs
    );

if( !$ok || defined($help_arg) ) {
    print "Usage: ./display_libcoll_tune_xml.pl [--help|-h] [--collective|-c COLLNAME] [--list|-l] XML_FILE\n";
    print "Example: ./display_libcoll_tune_xml.pl \$MPI_ROOT/etc/libcoll_tune_results.xml\n";
    exit 0;
}

if( scalar(@ARGV) != 1 && scalar(@ARGV) != 2 ) {
    print "Error: Must supply a libcoll_tune_results.xml file as the only argument\n";
    print "Usage: ./display_libcoll_tune_xml.pl [--help|-h] [--collective|-c COLLNAME] [--list|-l] XML_FILE\n";
    print "Example: ./display_libcoll_tune_xml.pl \$MPI_ROOT/etc/libcoll_tune_results.xml\n";
    exit -1;
}

my $tune_xml_file = $ARGV[0];
if( ! -e $tune_xml_file ) {
    print "Error: Cannot find the file specified:\n";
    print "\tFilename: $tune_xml_file\n";
    exit -2;
}

print "Processing XML: $tune_xml_file\n";
print "-"x70 . "\n";

my $xml = new XML::Simple();
my $data = $xml->XMLin($tune_xml_file);


#
# Extract the collective algorithm name/id for each collective op
#
extract_collective_names( $data->{collective_selection}->{collectives}->{collective} );

#
# Extract the selection table
#
extract_geometry_selection( $data->{collective_selection}->{geometries} );


# Display those values
my $collset;
my $collalg;
my $geoset;
my $geoitem;

foreach my $out_id (sort {$a <=> $b} keys %all_collectives) {
    my %algs_used;
    $collset = $all_collectives{$out_id};

    if( defined($show_coll) ) {
        if( lc($show_coll) ne lc($decode_coll_id{$collset->id()}) ) {
            next;
        }
    }
    print "-"x70 . "\n";
    print "Collective: ".$collset->id()." : ".$decode_coll_id{$collset->id()}."\n";
    print "-"x70 . "\n";
    
    # Display the 'key'
    foreach my $in_id (sort {$a <=> $b}  keys %{$collset->algs()} ) {
        $collalg = $collset->algs( $in_id );
        print "[".$collalg->id()."] ".$collalg->name()."\n";
    }

    if( defined($list_algs) ) {
        next;
    }

    # Display the chart
    # x = Bytes
    # y = NP
    my $min_np = 10000000;
    my $max_np = -1;
    my $min_bytes = 10000000;
    my $max_bytes = -1;
    my $cur_bytes = 1;

    # Find min/max
    foreach my $geo_size (sort {$a <=> $b}  keys %{$collset->geometry()} ) {
        if( $geo_size < $min_np ) {
            $min_np = $geo_size;
        }
        if( $geo_size > $max_np ) {
            $max_np = $geo_size;
        }

        $geoset = $collset->geometry($geo_size);
        foreach my $geoitem ( @{$geoset->items()} ) {
            if( $geoitem->size_min() < $min_bytes ) {
                $min_bytes = $geoitem->size_min();
            }
            if( $geoitem->size_max() > $max_bytes ) {
                $max_bytes = $geoitem->size_max();
            }
        }
    }

    # Display header
    printf("%4s | ", "");
    for( $cur_bytes = 1; $cur_bytes <= $max_bytes; $cur_bytes *= 2 ) {
        printf("%7d | ", $cur_bytes);
    }
    if( $min_bytes == $max_bytes && $max_bytes == 0 ) {
        printf("%7d | ", 0);
    }
    printf("\n");
    printf("-----+-", "");
    for( $cur_bytes = 1; $cur_bytes <= $max_bytes; $cur_bytes *= 2 ) {
        printf("--------+-", $cur_bytes);
    }
    if( $min_bytes == $max_bytes && $max_bytes == 0 ) {
        printf("--------+-", $cur_bytes);
    }
    printf("\n");

    #print "NP   : $min_np - $max_np\n";
    #print "Bytes: $min_bytes - $max_bytes\n";
    foreach my $geo_size (sort {$a <=> $b}  keys %{$collset->geometry()} ) {
        printf("%4d | ", $geo_size);
        $geoset = $collset->geometry($geo_size);

        if( $min_bytes == $max_bytes && $max_bytes == 0 ) {
            foreach my $geoitem ( @{$geoset->items()} ) {
                printf("%7d | ", $geoitem->algorithms(0));
            }
        }
        for( $cur_bytes = 1; $cur_bytes <= $max_bytes; $cur_bytes *= 2 ) {
            my $sanity_check = 0;
            foreach my $geoitem ( @{$geoset->items()} ) {
                if( $geoitem->size_min() <= $cur_bytes && $cur_bytes <= $geoitem->size_max() ) {
                    printf("%7d | ", $geoitem->algorithms(0));
                    if( exists $algs_used{$geoitem->algorithms(0)} ) {
                        $algs_used{$geoitem->algorithms(0)} += 1;
                    }
                    else {
                        $algs_used{$geoitem->algorithms(0)} = 1;
                    }
                    #printf("\t%7d - %7d ) %d\n",
                    #       $geoitem->size_min(),
                    #       $geoitem->size_max(),
                    #       $geoitem->algorithms(0) );
                    ++$sanity_check;
                    if( $sanity_check > 1 ) {
                        print "\nError\n";
                    }
                }
            }
        }
        print "\n";
    }

    printf("-----+-", "");
    for( $cur_bytes = 1; $cur_bytes <= $max_bytes; $cur_bytes *= 2 ) {
        printf("--------+-", $cur_bytes);
    }
    if( $min_bytes == $max_bytes && $max_bytes == 0 ) {
        printf("--------+-", $cur_bytes);
    }
    printf("\n");

    printf("Algorithms selected:\n");
    foreach my $akey ( sort {$algs_used{$b} <=> $algs_used{$a}} keys %algs_used ) {
        $collalg = $collset->algs( $akey );
        printf("\t(x%5d) [%2d] %15s\n", $algs_used{$akey}, $akey, $collalg->name());
    }
    printf("\n");
}
            

#print Dumper($data);

exit 0;

######################################################################
sub extract_collective_names( $ ) {
    my $alg_array_ref = shift(@_);
    my $collset;
    my $collalg;

    # Break out the XML into the structure format
    foreach my $ary_hash (@{$alg_array_ref}) {
        $collset = CollectiveSet->new();
        $collset->id( $ary_hash->{collective_id} );

        foreach my $key (keys $ary_hash->{algorithm}) {
            if( $key eq "name" ) {
                $collalg = CollectiveAlg->new();
                $collalg->id( $ary_hash->{algorithm}->{id} );
                $collalg->name( $ary_hash->{algorithm}->{$key} );

                $collset->algs( $collalg->id() , $collalg );
            }
            elsif( $key eq "id" ) {
                ;
            }
            else {
                $collalg = CollectiveAlg->new();
                $collalg->id( $ary_hash->{algorithm}->{$key}->{id} );
                $collalg->name( $key );

                $collset->algs( $collalg->id() , $collalg );
            }
        }
        $all_collectives{ $collset->id() } = $collset;
    }

    return 0;
}

sub extract_geometry_selection( $ ) {
    my $data_geometries = shift(@_);

    print "PPN: " . $data_geometries->{ppn} . "\n";

    my $geo_ref = $data_geometries->{geometry};

    # For each Geometry size
    foreach my $item (@{$geo_ref}) {
        #print "-"x30 . " Geometry Size: ".$item->{geometry_size}."\n";
        my $coll_ref = $item->{coll};

        # For each collective
        foreach my $coll (@{$coll_ref}) {
            #print "\t" . "-"x30 . " Collective ID: ".$coll->{coll_id}."\n";
            my $collset = $all_collectives{$coll->{coll_id}};
            my $geoset;
            my $geoitem;
            # See if we already have a geometry set for this
            #my $tref = $collset->geometry();
            if( exists $collset->geometry()->{$item->{geometry_size}} ) {
                $geoset = $collset->geometry($item->{geometry_size});
            }
            else {
                $geoset = GeometrySelection->new();
                $geoset->size( $item->{geometry_size} );
                $collset->geometry( $geoset->size(), $geoset );
            }

            # For each datapoint in that collective for this geometry
            if( ref($coll->{message}) eq "ARRAY" ) {
                my $message_ref = $coll->{message};
                foreach my $msg (@{$message_ref}) {
                    $geoitem = GeometryItem->new();
                    $geoitem->size_min( $msg->{size_min} );
                    $geoitem->size_max( $msg->{size_max} );
                    @{$geoitem->algorithms} = split(/\s*,\s*/, $msg->{algorithms});

                    push( @{$geoset->items()}, $geoitem);
 
                    #print "\t\t" . "-"x30 . ": ...\n";
                    #printf("%7d - %7d ) %s\n",
                    #       $msg->{size_min}, $msg->{size_max},
                    #       $msg->{algorithms});
                }
            }
            if( ref($coll->{message}) eq "HASH" ) {
                my $msg = $coll->{message};

                $geoitem = GeometryItem->new();
                $geoitem->size_min( $msg->{size_min} );
                $geoitem->size_max( $msg->{size_max} );
                @{$geoitem->algorithms} = split(/\s*,\s*/, $msg->{algorithms});

                push( @{$geoset->items()}, $geoitem);

                #printf("%7d - %7d ) %s\n",
                #       $msg->{size_min}, $msg->{size_max},
                #       $msg->{algorithms});
            }
        }
    }

    return 0;
}
