/*
 *   Copyright (c) 1997-2016 IBM Corporation.  All rights reserved.
 *
 *	Function:	- example: thread-safe MPI
 *
 *      Algorithm:	- N clients loop MAX_WORK times doing work. As part 
 *      		  of a single work item, a client must request service 
 *			  from one of N servers at random. The server keeps a 
 *		          count of the requests handled and prints a log
 *			  of the requests to stdout. Once all the clients are 
 *			  done working, the servers are shutdown.
 */

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <pthread.h>

#define	MAX_WORK	40
#define SERVER_TAG	88
#define CLIENT_TAG	99

#define REQ_SHUTDOWN 	-1

static	int		service_cnt = 0;

int
process_request(int request)
{
	if (request != REQ_SHUTDOWN) service_cnt++;
	return request;
}

void*
server(void *args)
{
	int		rank, request;
	MPI_Status	status;

	rank = *((int*)args);

	while (1) {
		MPI_Recv(&request, 1, MPI_INT, MPI_ANY_SOURCE,
					SERVER_TAG, MPI_COMM_WORLD, &status);

		if (process_request(request) == REQ_SHUTDOWN)
			break;

		MPI_Send(&rank, 1, MPI_INT, status.MPI_SOURCE,
					CLIENT_TAG, MPI_COMM_WORLD);

		printf("server [%d]: processed request %d for client %d\n",
					rank, request, status.MPI_SOURCE);
	}

	printf("server [%d]: total service requests: %d\n", rank, service_cnt);

	return (void*) 0;
}

void
client(int rank, int size)
{
	int		w, server, ack;
	MPI_Status	status;

	for (w = 0; w < MAX_WORK; w++) {
		server = rand()%size;

		MPI_Sendrecv(&rank, 1, MPI_INT, server, SERVER_TAG,
				&ack, 1, MPI_INT, server, CLIENT_TAG,
				MPI_COMM_WORLD, &status);

		if (ack != server) {
			printf("server failed to process my request\n");
			MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
		}
	}
}

void
shutdown_servers(int rank)
{
	int	request_shutdown = REQ_SHUTDOWN;

	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Send(&request_shutdown, 1, MPI_INT, 
			rank, SERVER_TAG, MPI_COMM_WORLD);
}

int
main(int argc, char *argv[])
{
	int		rank, size, rtn;
	pthread_t	mtid;
	MPI_Status	status;
	int  		my_value, his_value;
	int      thread_lvl;

	MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &thread_lvl);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

        rtn = pthread_create(&mtid, 0, server, (void*)&rank);
        if (rtn != 0) {
       		printf("pthread_create failed\n");
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
	}

	client(rank, size);
	shutdown_servers(rank);

    	rtn = pthread_join(mtid, 0);
        if (rtn != 0) {
       		printf("pthread_join failed\n");
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OTHER);
	}

	MPI_Finalize();
	exit(0);
}
